import { Injectable } from "@angular/core";
import { BehaviorSubject, from } from "rxjs";
import { HttpClient } from "@angular/common/http";
import { map } from "rxjs/operators";
import { Capacitor } from "@capacitor/core";
import "@capacitor-community/sqlite";
import {
  CapacitorSQLite,
  SQLiteDBConnection,
  SQLiteConnection,
  capSQLiteSet,
  capSQLiteChanges,
  capEchoResult,
  capSQLiteResult,
  capSQLiteValues,
  JsonSQLite,
} from "@capacitor-community/sqlite";
import { Storage } from "@ionic/storage-angular";

import { Note } from "./note.model";
import { NotificationsService } from "./notifications.service";

const DB_NAME_KEY = "db-name";
const DB_SETUP_KEY = "db-setup";

@Injectable({
  providedIn: "root",
})
export class StorageService {
  dbReady = new BehaviorSubject(false);
  dbName = "";
  sqlite: SQLiteConnection;
  db: SQLiteDBConnection;
  // dbOptions: capSQLiteOptions = { database: this.dbName };
  isService: boolean = false;
  platform: string;
  native: boolean = false;
  private _storage: Storage | null = null;

  constructor(
    private http: HttpClient,
    private storage: Storage,
    private notificationsService: NotificationsService
  ) {}

  /**
   * Plugin Initialization
   */
  async initializePlugin(): Promise<boolean> {
    return new Promise(async (resolve) => {
      const storage = await this.storage.create();
      console.log(storage);
      this._storage = storage;
      console.log(this._storage);
      await this._storage.set('test', 'test123value');
      const dbKey = await this._storage.get('test');
      console.log(dbKey);
      console.log("starting sqlite initialization...");
      this.platform = Capacitor.getPlatform();
      console.log("*** platform " + this.platform);
      if (this.platform === "ios" || this.platform === "android") {
        this.native = true;
      }
      const sqlitePlugin: any = CapacitorSQLite;
      this.sqlite = new SQLiteConnection(sqlitePlugin);
      this.isService = true;
      console.log("$$$ in service this.isService " + this.isService + " $$$");
      this.setupDatabase();
      resolve(true);
    });
  }
  /**
   * Echo a value
   * @param value
   */
  async echo(value: string): Promise<capEchoResult> {
    console.log("&&&& in echo this.sqlite " + this.sqlite + " &&&&");
    if (this.sqlite != null) {
      return await this.sqlite.echo(value);
    } else {
      return null;
    }
  }

  private async setupDatabase() {
    console.log("setting up database");
    try {
      const dbSetupDone = await this._storage.get(DB_SETUP_KEY);
      console.log(dbSetupDone);
      if (!dbSetupDone) {
        console.log('no db here');
        this.downloadDatabase();
      } else {
        console.log('db already setup');
        this.dbName = await this._storage.get(DB_NAME_KEY);
        let db = await this.createConnection();
        if (db) {
          this.dbReady.next(true);
        }
      }
    } catch (e) {
      console.error(e);
      console.log("error in getting dbkey");
    }
  }

  private async createConnection(): Promise<SQLiteDBConnection | null> {
    console.log("setting up db connection");
    try {
      if (this.sqlite != null) {
        let isConn = false;
        const retCC = (await this.sqlite.checkConnectionsConsistency()).result;
        if (retCC) {
          isConn = (await this.sqlite.isConnection(this.dbName, false)).result;
        }
        console.log(retCC);
        console.log(isConn);
        if(retCC && isConn) {
          this.db = await this.sqlite.retrieveConnection(this.dbName, false);
        } else {
          this.db = await this.sqlite.createConnection(
            this.dbName,
            false,
            "no-encryption",
            1,
            false
          );
        }
        console.log(this.db);
        if (this.db != null) {
          await this.db.open();
          return this.db;
        } else {
          return null;
        }
      } else {
        return null;
      }
    } catch (e) {
      console.error('error in checking sq lite connection');
      console.error(e);
    }
  }

  private downloadDatabase(update = false) {
    console.log("downloading database");
    this.http
      .get("../assets/start_db.json")
      .subscribe(async (jsonExport: JsonSQLite) => {
        const jsonstring = JSON.stringify(jsonExport);
        let isValid = false;
        try {
          const isValidTry = await CapacitorSQLite.isJsonValid({ jsonstring });
          console.log(isValidTry);
          isValid = isValidTry.result;
        } catch (e) {
          console.error(e);
        }
        console.log(isValid);
        if (isValid) {
          console.log("dbjson is valid");
          this.dbName = jsonExport.database;
          try {
            await this._storage.set(DB_NAME_KEY, this.dbName);
            await this.sqlite.importFromJson(jsonstring);
            await this._storage.set(DB_SETUP_KEY, "1");
          } catch (e) {
            console.error(e);
          }
          console.log("db set");
          // Your potential logic to detect offline changes later
          let db = await this.createConnection();
          if (db) {
            this.dbReady.next(true);
            if (!update) {
              await db.createSyncTable();
            } else {
              await db.setSyncDate("" + new Date().getTime());
            }
          }
        }
      });
  }

  async getTable(table: string) {
    console.log("getTable " + table + " from " + this.dbName);
    const statement = "SELECT * FROM " + table;
    return await this.db.query(statement);
    /*
    return this.dbReady.pipe(
      switchMap(async (isReady) => {
          if (isReady) {
            const statement = "SELECT * FROM " + table;
            return await from(this.db.query(statement));
          } else {
            return of({ values: [] });
          }
      })
    );*/
  }

  async getNoteById(noteId) {
    const statement = "SELECT * FROM notes WHERE notes.id=" + noteId + " ;";
    return (await this.db.query(statement)).values[0];
  }

  async moveNote(origin, target, note: Note) {
    // const moveStatement = 'INSERT INTO trash (title, note) SELECT (title, note) FROM notes WHERE notes.id=' + noteId + ';';
    const delStatement =
      "DELETE FROM " + origin + " WHERE id = " + note.id + ";";
    const moveStatement =
      "INSERT INTO " +
      target +
      " (title, note, color, created, last_modified) VALUES (?, ?, ?, ?, ?);";
    const values = [
      note.title,
      note.note,
      note.color,
      note.created,
      note.last_modified,
    ];
    await this.db.run(moveStatement, values);
    const remindCheckRes = await this.notificationsService.checkForReminder(
      note.id
    );
    console.log("remindercheck before deleting note:");
    console.log(remindCheckRes);
    if (remindCheckRes) {
      console.log("reminder found");
      this.notificationsService.cancel(note.id);
    }
    return this.db.execute(delStatement);
  }

  async clearArchive() {
    const moveStatement =
      "INSERT INTO trash (title, note, color, created, last_modified) SELECT title, note, color, created, last_modified FROM archive;";
    const delStatement = "DELETE FROM 'archive';";
    const moveRes = await this.db.execute(moveStatement);
    console.log("moveResult.changes.changes ", moveRes.changes.changes);
    return this.db.execute(delStatement);
  }

  clearTrash() {
    const statement = "DELETE FROM 'trash';";
    return this.db.execute(statement);
  }

  updateNote(noteId, title: string, notetext: string, color: string) {
    const statement =
      "UPDATE notes SET title = '" +
      title +
      "', note = '" +
      notetext +
      "', color = '" +
      color +
      "', last_modified = '" +
      Date.now() +
      "' WHERE id = '" +
      noteId +
      "';";
    console.log(statement);
    return this.db.execute(statement);
  }

  async addNote(title: string, notetext: string, color: string) {
    const statement =
      "INSERT INTO notes (title, note, color, created, last_modified) VALUES (?, ?, ?, ?, ?);";
    const values = [title, notetext, color, Date.now(), Date.now()];
    console.log(statement);
    return this.db.run(statement, values);
  }

  deleteNoteFromTrash(noteId: number) {
    const statement = "DELETE FROM trash WHERE id = " + noteId + ";";
    return this.db.execute(statement);
  }

  async getDatabaseExport(mode: string) {
    const dbName = await this.storage.get(DB_NAME_KEY);
    // await Storage.set({ key: DB_SETUP_KEY, value: null });
    return this.db.exportToJson(mode);
  }

  async importDatabase(jsonImport) {
    this.dbReady.next(false);
    this.sqlite.closeConnection(this.dbName, false);
    console.log('JSON IMPORT DATA:');
    console.log(jsonImport);
    try {
      console.log("trying import...");
      const jsonstring = JSON.stringify(jsonImport.export);
      const isValidTry = await this.sqlite.isJsonValid(jsonstring);
      if (isValidTry.result) {
        console.log("dbjson is valid");
        try {
          await this.sqlite.importFromJson(jsonstring);
          let db = await this.createConnection();
          if (db) {
            this.dbReady.next(true);
        }
    } catch (e) {
      console.log(e);
    }
  }
} catch (e) {
  console.log(e);
}
  }

  async resetDb() {
    this.dbReady.next(false);
    await this.deleteDatabase();
    await this.clearLocalStorageData();
    await this.setupDatabase();
    /*this.http
      .get("../assets/start_db.json")
      .subscribe(async (jsonExport: JsonSQLite) => {
        const jsonstring = JSON.stringify(jsonExport);
        let isValid = false;
        this.clearLocalStorageData();
        try {
          const isValidTry = await CapacitorSQLite.isJsonValid({ jsonstring });
          console.log(isValidTry);
          isValid = isValidTry.result;
        } catch (e) {
          console.error(e);
        }
        console.log(isValid);
        if (isValid) {
          console.log("dbjson is valid");
          this.dbName = jsonExport.database;
          try {
            await this._storage.set(DB_NAME_KEY, this.dbName);
            await this.sqlite.importFromJson(jsonstring);
            await this._storage.set(DB_SETUP_KEY, "1");
            this.dbReady.next(true);
          } catch (e) {
            console.error(e);
          }
          console.log("db reset");
        }
      });*/
  }

  async deleteDatabase() {
    const dbName = await this._storage.get(DB_NAME_KEY);
    await this._storage.set(DB_SETUP_KEY, null);
    return CapacitorSQLite.deleteDatabase({ database: dbName });
  }

  saveLocalStorageData(darkMode: boolean) {
    const userData = JSON.stringify({ darkMode });
    this._storage.set("vioNoteLocalData", userData);
  }

  private clearLocalStorageData() {
    console.log('local Storage cleared');
    this._storage.remove("vioNoteLocalData");
  }

  getLocalStorageData() {
    return from(this._storage.get("vioNoteLocalData")).pipe(
      map((storedData) => {
        if (!storedData) {
          return null;
        }
        const parsedData = JSON.parse(storedData) as {
          darkMode: boolean;
        };
        return {
          darkMode: parsedData.darkMode,
        };
      })
    );
  }
}
